#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pri.h"


/* declare a type and a size of the data buffer */
#define BUFFERSIZE  128
struct iobuffer {
    unsigned char data[BUFFERSIZE];
    size_t size;        /* number of bytes in the "data" buffer */
    int index;          /* current index in the "data" buffer */
    FILE *file;         /* file from which to read */
};


/* simulate a byte streaming interface with macros */
#define GETBYTE(buffer)       getbyte((buffer))
#define PUTBYTE(output,byte)  fputc((byte),(output))

int getbyte(struct iobuffer *buffer); /* forward declaration, for unrle() */

#include "unrle.c"



/* getbuffer()
 * Helper function to read a new input buffer, of the constant size BUFFERSIZE.
 * The function returns 1 on success and 0 on failure (EOF).
 */
int getbuffer(struct iobuffer *buffer)
{
    buffer->size = fread(buffer->data, 1, BUFFERSIZE, buffer->file);
    buffer->index = 0;
    return buffer->size > 0;
}

/* getbyte()
 * Helper function to get a single byte from the input file. This byte comes
 * from the buffer; if the buffer is empty, new data is read from the file.
 * On EOF (and if the buffer is empty), the function returns EOF;
 */
int getbyte(struct iobuffer *buffer)
{
    if (buffer->index >= buffer->size)
        getbuffer(buffer);
    return (buffer->index < buffer->size) ? buffer->data[buffer->index++] : EOF;
}

/* pri_readheader()
 * Helper function that reads a bitmap header byte-for-byte and optionally
 * aligns the multi-byte fields to little Endian
 */
int pri_readheader(PRHDR *pri, struct iobuffer *buffer)
{
    /* we assume that the buffer is at least the size of the PRI bitmap header;
     * for smaller buffers, it may not make sense to use a buffering scheme
     */
    #if BUFFERSIZE < PRHEADERSIZE
        #error BUFFERSIZE is declared too small
    #endif
    int count;

    for (count = 0; count < PRHEADERSIZE; count++)
        GETBYTE(buffer);

    pri->size   = PRHDR_SIZE(buffer->data);
    pri->id     = PRHDR_ID(buffer->data);
    pri->layout = PRHDR_LAYOUT(buffer->data);
    pri->width  = PRHDR_WIDTH(buffer->data);
    pri->height = PRHDR_HEIGHT(buffer->data);
    pri->depth  = PRHDR_DEPTH(buffer->data);

    if (pri->size == 0 || pri->id != PRSIGNATURE)
        return 0;

    return 1;
}

/* write_ppm()
 * Writes the header of a PBM image, then copies pixels from input to output.
 * For a PBM file, 0 is white and 1 is black; for a PRI file, the inverse is
 * true. So, we invert all bytes before writing them.
 */
void write_ppm(FILE *output, int width, int height, struct iobuffer *buffer)
{
    struct rlestat stat = { 0 };
    long size = height * ((width + 7) / 8);
    int c;

    fprintf(output, "P4\n%d %d\n", width, height);

    while (size-- > 0) {
        c = unrle(buffer, &stat);
        PUTBYTE(output, ~c);
    } /* while */
}

int main(int argc, char *argv[])
{
    FILE *output;
    PRHDR pri;

    /* declare a buffer for the input, make sure to zero-initialize the
     * structure
     */
    struct iobuffer in_buffer = { 0 };

    /* open files */
    in_buffer.file = (argc >= 2) ? fopen(argv[1], "rb") : stdin;
    output = (argc >= 3) ? fopen(argv[2], "wb") : stdout;
    if (in_buffer.file == NULL) {
        perror("Input file");
        return 1;
    }
    if (output == NULL) {
        perror("Output file");
        return 1;
    }

    /* stream in a header */
    for ( ;; ) {
        if (!pri_readheader(&pri, &in_buffer)) {
            fprintf(stderr, "No supported bitmap type found\n");
            return 1;
        }
        if (pri.layout == 0 && pri.depth == 1)
            break;
        /* ignore this bitmap (without going through every byte) */
        while (pri.size > PRHEADERSIZE && GETBYTE(&in_buffer) != EOF)
            pri.size--;
    }

    /* write the PPM header */
    write_ppm(output, pri.width, pri.height, &in_buffer);

    /* clean up */
    if (in_buffer.file != stdin)
        fclose(in_buffer.file);
    if (output != stdout)
        fclose(output);

    return 0;
}
